
#include "Watch.h"

uint16_t clockCircleColors[] = { ILI9341_RED, ILI9341_YELLOW, ILI9341_AQUA, ILI9341_CRIMSON, ILI9341_BLUEVIOLET };
char daysOfTheWeek[7][12] = { "Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday" };
char _Month[12][12] =  { "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December" };

void WatchClass::init(Adafruit_GFX *tft, uint16_t BackColor, uint16_t TextColor,bool FullMonth,uint16_t x, uint16_t y)
{
	_tft = tft;
	_BackColor = BackColor;
	_TextColor = TextColor;
	 
	_clockCenterX = x;
	_clockCenterY = y;

	ChangedTime.Hour = true;
	ChangedTime.Minute = true;
	ChangedTime.Second = true;

	ChangedTime.Day = true;
	ChangedTime.Month = true;
	ChangedTime.Year = true;
	ChangedTime.Wday = true;

	_FullMonth = FullMonth;
	_ScreenHeight = _tft->height(); 
	_ScreenWidth =  _tft->width();
	 
}


void  WatchClass::initAnalogColors(uint16_t HourHand, uint16_t MinHand, uint16_t SecHand, uint16_t Background)
{
 _HourHand = HourHand;
 _MinHand = MinHand;
 _SecHand = SecHand;
 _Background = Background;
}


void WatchClass::drawCircle(int x0, int y0, int radius, uint16_t color)
{
	int x = radius;
	int y = 0;
	int radiusError = 1 - x;
	Serial.print("Kreisfarbe ");
	Serial.println(color);
	while (x >= y)
	{
		//drawPixel(x,y,color);//set pixel (x,y) to specified color. This is slow because we need to send commands setting the x and y, then send the pixel data.
		_tft->drawPixel(x + x0, y + y0, color);
		_tft->drawPixel(y + x0, x + y0, color);
		_tft->drawPixel(-x + x0, y + y0, color);
		_tft->drawPixel(-y + x0, x + y0, color);
		_tft->drawPixel(-x + x0, -y + y0, color);
		_tft->drawPixel(-y + x0, -x + y0, color);
		_tft->drawPixel(x + x0, -y + y0, color);
		_tft->drawPixel(y + x0, -x + y0, color);
		y++;
		if (radiusError < 0)
		{
			radiusError += 2 * y + 1;
		}
		else
		{
			x--;
			radiusError += 2 * (y - x) + 1;
		}
	}
}

void WatchClass::drawHand(int xStart, int yStart, int angle, int radius, uint16_t color)

// xStart, yStart are starting point of hand (center of clock face)
// angle is location of hand on dial (0-60)
// radius is length of hand
{
	// Sines for hand positions from 0 to 15 minutes. All of the other
	// hand position sines can be determined by flipping x and y. These
	// are all multiplied by 256 so that we can use integers to store them.
	// Cosines can be also be determined by taking the sine of the angle minus
	// 15 degrees and flipping x and y.
	static int sine[16] = { 0, 27, 54, 79, 104, 128, 150, 171, 190, 201, 221, 233, 243, 250, 254, 255 };
	int xEnd, yEnd, quadrant, x_flip, y_flip;

	// calculate which quadrant the hand lies in
	quadrant = angle / 15;

	// We could have a bigger table of sine values that handle every hand position from 0 to 60,
	// but we can avoid using that memory by doing some quick reflections and rotations.
	switch (quadrant)
	{
	case 0: x_flip = 1; y_flip = -1; break;
	case 1: angle = abs(angle - 30); x_flip = y_flip = 1; break;
	case 2: angle = angle - 30; x_flip = -1; y_flip = 1; break;
	case 3: angle = abs(angle - 60); x_flip = y_flip = -1; break;
	default:  x_flip = y_flip = 1; // this should not happen
	}

	xEnd = xStart;
	yEnd = yStart;
	// OK, here's the tricky part. Look up the sine for the hand
	// angle, then multiply it by the hand length. Then, because
	// our sine table values are all pre-multiplied by 256 so we
	// can store them as integers, we need to divide by 256 to
	// correct for that. Which is, in integer math, the same as
	// saying that we'll shift the values right by 8 bit positions.
	// Doing a shift-right is faster than doing an integer divide by 256
	// (the compiler is probably smart enough to turn that
	// divide by 256 into a right shift anyway, but let's do
	// an explicit shift because it's cool).
	//
	// The equivalent floating-point call would look something like:
	// xEnd = sine(angle) * radius;
	xEnd += (x_flip * ((sine[angle] * radius) >> 8));
	yEnd += (y_flip * ((sine[15 - angle] * radius) >> 8));

	_tft->drawLine(xStart, yStart, xEnd, yEnd, color);

}

void WatchClass::drawAdvancedCircle(void)
{
	for (int i = 0; i < 5; i++)
	{
		drawCircle(_clockCenterX, _clockCenterY, clockCircleRadius + i, clockCircleColors[i]);
	}
}


void WatchClass::SetAnalogTime(tmElements_t Value)
{
	int hourAngle;
	if (Value.Hour >= 12)
		hourAngle = Value.Hour - 12;
	else
		hourAngle = Value.Hour;
	hourAngle *= 5;
	// To give the hour hand a real
	// analog-clock feel, it shouldn't point
	// straight at the hour, but should move between
	// the hours as the minute hand sweeps around.
	// So, let's nudge it forward a smidge by
	// the minutes scaled down to the space within 1 hour, or
	// divide them by 12.
	hourAngle += Value.Minute / 12;
	drawHand(_clockCenterX, _clockCenterY, hourAngle, clockHourHandLength, _HourHand);
	drawHand(_clockCenterX, _clockCenterY, Value.Minute, clockMinuteHandLength, _MinHand);
	drawHand(_clockCenterX, _clockCenterY, Value.Second, clockSecondHandLength, _SecHand);
}

void WatchClass::ClearAnalogTime(tmElements_t Value)
{
	int hourAngle;
	if (Value.Hour >= 12)
		hourAngle = Value.Hour - 12;
	else
		hourAngle = Value.Hour;
	hourAngle *= 5;
	// To give the hour hand a real
	// analog-clock feel, it shouldn't point
	// straight at the hour, but should move between
	// the hours as the minute hand sweeps around.
	// So, let's nudge it forward a smidge by
	// the minutes scaled down to the space within 1 hour, or
	// divide them by 12.
	hourAngle += Value.Minute / 12;


	drawHand(_clockCenterX, _clockCenterY, hourAngle, clockHourHandLength, _Background);
	drawHand(_clockCenterX, _clockCenterY, Value.Minute, clockMinuteHandLength, _Background);
	drawHand(_clockCenterX, _clockCenterY, Value.Second, clockSecondHandLength,  _Background);

	_tft->fillCircle(_clockCenterX, _clockCenterY, 5, ILI9341_GOLD);
	_tft->fillCircle(_clockCenterX, _clockCenterY, 3, ILI9341_NAVY);
}


void WatchClass::SetDigiTime(tmElements_t OldValue, tmElements_t NewValue, bool OnOFF)
{
	TimeHasChanged_t ChangedTime;
	ChangedTime = SetTimeBool(OldValue, NewValue);
	
	sprintf(Buffer, "H%d M%d S%d ", ChangedTime.Hour, ChangedTime.Minute, ChangedTime.Second);
	Serial.println(Buffer);

	// Print date in EU format DD:MM:YY (Switch the order in which day, month, year that you like to use)
	
	 if (ChangedTime.Hour)
	 {
	  // Hour Back
	  _tft->setTextColor(_BackColor);
	  _tft->setCursor(clockPosHour, _ScreenHeight - clockYFrame); //Set the cursor where you want to start printing the time
	   if (OldValue.Hour < 10) _tft->print(0); //print a leading 0 if hour value is less than 0
	  _tft->print(OldValue.Hour);
	  _tft->print(":");

	  // Hour
	  _tft->setTextColor(_TextColor);
	  _tft->setCursor(clockPosHour, _ScreenHeight - clockYFrame); //Set the cursor where you want to start printing the time
	  if (NewValue.Hour < 10) _tft->print(0); //print a leading 0 if hour value is less than 0
	  _tft->print(NewValue.Hour);
	  _tft->print(":");
	 }

	 if (ChangedTime.Minute)
	 {
	// Minute Back
	_tft->setTextColor(_BackColor);
	_tft->setCursor(clockPosMin, _ScreenHeight - clockYFrame); //Set the cursor where you want to start printing the time
	if (OldValue.Minute < 10) _tft->print(0); //print a leading 0 if minute value is less than 0
	_tft->print(OldValue.Minute);
	_tft->print(":");
	// Minute 
	_tft->setTextColor(_TextColor);
	_tft->setCursor(clockPosMin, _ScreenHeight - clockYFrame); //Set the cursor where you want to start printing the time
	if (NewValue.Minute < 10) _tft->print(0); //print a leading 0 if minute value is less than 0
	_tft->print(NewValue.Minute);
	_tft->print(":");
	 }


	 if (ChangedTime.Second)
	 {
	// Seconds Back
	_tft->setTextColor(_BackColor);
	_tft->setCursor(clockPosSec, _ScreenHeight - clockYFrame); //Set the cursor where you want to start printing the time
	if (OldValue.Second < 10) _tft->print(0); //print a leading 0 if seconds value is less than 0
	_tft->print(OldValue.Second);
	//_tft->print(" "); //just a empty space after the seconds
	sprintf(Buffer, "SOld %d SNew %d", OldValue.Second, NewValue.Second);

	Serial.println(Buffer);
		// Seconds
	_tft->setTextColor(_TextColor);
	_tft->setCursor(clockPosSec, _ScreenHeight - clockYFrame); //Set the cursor where you want to start printing the time
	if (NewValue.Second < 10) _tft->print(0); //print a leading 0 if seconds value is less than 0
	_tft->print(NewValue.Second);
//	_tft->print(" "); //just a empty space after the seconds
	}

	 /// Set Date Part

	 if (ChangedTime.Day)
	 {
		 // Day Back
		 _tft->setTextColor(_BackColor);
		 _tft->setCursor(clockPosDay, _ScreenHeight - clockYFrame);
		 if (OldValue.Day < 10) _tft->print(0);
		 _tft->print(OldValue.Day);
		 _tft->print(".");
	

		 // Day
		 _tft->setTextColor(_TextColor);
		 _tft->setCursor(clockPosDay, _ScreenHeight - clockYFrame);
		 if (NewValue.Day < 10) _tft->print(0);
		 _tft->print(NewValue.Day);
		 _tft->print(".");
		 }

	 

	 if (ChangedTime.Month)
	 {
		 /// Month Back
		 _tft->setTextColor(_BackColor);
		 _tft->setCursor(clockPosMonth, _ScreenHeight - clockYFrame);
		 if (_FullMonth)
		 {
			 strcpy(Buffer, _Month[OldValue.Month]);
			 _tft->print(Buffer);
		 }
		 else
		 {
			 if (OldValue.Month < 10) _tft->print(0);
			 _tft->print(OldValue.Day);
		 }
		 _tft->print(".");


		 // Month
		 _tft->setTextColor(_TextColor);
		 _tft->setCursor(clockPosMonth, _ScreenHeight - clockYFrame);
		 if (_FullMonth)
		 {
			 strcpy(Buffer, _Month[OldValue.Month]);
			_tft->print(Buffer);
		 }
		 else
		 {
			 if (NewValue.Month < 10) _tft->print(0);
			 _tft->print(NewValue.Day);
		 }
		 _tft->print(".");
	 }

	

	 if (ChangedTime.Year)
	 {
		 // Year Back
		 uint8_t Distance = 0;
		 if (_FullMonth)
			 Distance = 43;
		 
		 _tft->setTextColor(_BackColor);
		 _tft->setCursor(clockPosYear + Distance, _ScreenHeight - clockYFrame);
		 _tft->print(OldValue.Year + 1900);

		 // Year
		 _tft->setTextColor(_TextColor);
		 _tft->setCursor(clockPosYear + Distance, _ScreenHeight - clockYFrame);
		 _tft->print(NewValue.Year + 1900);
	 }


	 // Day of the Week !!!!
	 
	 if (ChangedTime.Wday)
	 {
		 strcpy(Buffer, daysOfTheWeek[OldValue.Wday]);
		// wkday[3] = ' '; wkday[4] = '\0';
		 _tft->setTextColor(_BackColor);
		 _tft->setCursor(15, _ScreenHeight - clockYFrame); //Set the cursor where you want to start printing the date
		 _tft->print(Buffer); // To keep the text compact, only print the first 3 letters.

		 strcpy(Buffer, daysOfTheWeek[NewValue.Wday]);
		// wkday[3] = ' '; wkday[4] = '\0';
		 _tft->setTextColor(_TextColor);
		 _tft->setCursor(15, _ScreenHeight - clockYFrame); //Set the cursor where you want to start printing the date
		 _tft->print(Buffer); // To keep the text compact, only print the first 3 letters.
	 }
	 
	_tft->drawRoundRect(0, 240 - 20, 320, 20, 10, ILI9341_BLANCHEDALMOND);
}



TimeHasChanged_t WatchClass::SetTimeBool(tmElements_t NewValue, tmElements_t OldValue)
{
	TimeHasChanged_t Result;
   Result.Day = false;
   Result.Month = false;
   Result.Year = false;

	Result.Hour = false;
	Result.Minute = false;
	Result.Second = false;

	if (NewValue.Hour != OldValue.Hour)
		Result.Hour = true;

	if (NewValue.Minute != OldValue.Minute)
		Result.Minute = true;
	
	if (NewValue.Second != OldValue.Second)
		Result.Second = true;

	if (NewValue.Day != OldValue.Day)
		Result.Day = true;

	if (NewValue.Month != OldValue.Month)
		Result.Month = true;

	if (NewValue.Year != OldValue.Year)
		Result.Year = true;

	if (NewValue.Wday != OldValue.Wday)
		Result.Wday = true;

	return Result;
}



WatchClass Watch;

